'**********************************************************'
'* --< ADVANCED GRAPHICS INTERFACE - BY TENIE REMMEL >--  *'
'*                                                        *'
'* AGI.ASM - This is a highly optimized graphics library  *'
'* for the undocumented 640x400x256 X mode.  It is quite  *'
'* complete, and is also small (4K).  This is intended    *'
'* for Tiny, Small and Compact model C and asm programs.  *'
'*                                                        *'
'* This library is public domain.  However, if you intend *'
'* to distribute this source code, you MUST retain this   *'
'* header in its original form.                           *'
'**********************************************************'

'**********************************************************'
'* Adapted for PowerBasic 3.2                             *'
'* Hans Lunsing, january 16, 1997                         *'
'* e-mail jlunsing@doge.nl                                *'
'**********************************************************'

'-----------------------------------------------------------------------
$COMPILE        UNIT

'Compiler directives for code generation

$CPU            80286   'Generate code for processor {8086 | 80286 | 80386}
$DIM            ALL     'Declararations required for {ALL | ARRAY | NONE}
$OPTION SIGNED  OFF     'Return addresses by VARSEG etc. as signed integers

'Attention: the assembler code is _not_ suited for 8086 CPU's.
'It doesn't help to set $CPU to 8086.

'-----------------------------------------------------------------------
$INCLUDE        "pbagi.bi"
$INCLUDE        "opcod286.inc"

$CODE SEG       "pbagi_text"

'**************************** Global variables

FontPtr:
asm             dd      0
FontWidth:
asm             db      0
FontHeight:
asm             db      0

MinX:
asm             dw      0
MinY:
asm             dw      0
MaxX:
asm             dw      0
MaxY:
asm             dw      0
Colour:
asm             dw      0

PTable:
asm             db      1
asm             db      2
asm             db      4
asm             db      8

'**************************** InitGraph variables

CRTCdata:                                       'CRTC parameters
asm             dw      &h0E11
asm             dw      &hBF06
asm             dw      &h1F07
asm             dw      &h4009
asm             dw      &h9C10
asm             dw      &h8E11
asm             dw      &h8F12
asm             dw      &h0014
asm             dw      &h9615
asm             dw      &hB916
asm             dw      &hE317

'**************************** Line variables

deltav_str:
asm             dw      0                       ;Deviation change straight
deltav_diag:
asm             dw      0                       ;Deviation change diagonal
xinc_str:
asm             dw      0                       ;X increment straight
xinc_diag:
asm             dw      0                       ;X increment diagonal
yinc_str:
asm             dw      0                       ;Y increment straight
yinc_diag:
asm             dw      0                       ;Y increment diagonal
scratch:
asm             dw      0                       ;Scratch variables
asm             dw      0

'**************************** OutChar variables

init_mask:
asm             db      0
asm             db      0                       ;extra for even addresses

'**************************** GetImage variables

gi_width:
asm             dw      0                       ;Total width
gi_cplane:
asm             dw      0                       ;Current plane
gi_plane:
asm             db      0                       ;Starting plane
asm             db      0                       ;extra for even addresses

'**************************** PutImage variables

pi_width:
asm             dw      0                       ;Total width
pi_cplane:
asm             dw      0                       ;Current plane
pi_plane:
asm             db      0                       ;Starting plane
asm             db      0                       ;extra for even addresses

'**************************** Circle variables

xsq:
asm             dw      0                       ;X variable
asm             dw      0
ysq:
asm             dw      0                       ;Y variable
asm             dw      0
rsq:
asm             dw      0                       ;R variable
asm             dw      0

'**************************** Ellipse variables

e_xnum:
asm             dw      0
e_ynum:
asm             dw      0
e_denom:
asm             dw      0

'**************************** Fillarea variables

stack_top:
asm             dw      0                       ;Stack top
old_color:
asm             dw      0                       ;Old color
new_color:
asm             dw      0                       ;New color

'-----------------------------------------------------------------------

FUNCTION InitGraph () PUBLIC AS INTEGER

' InitGraph returns 0 if the 640x400x256 mode couldn't be initialized
' (for instance with a non-VESA compatible video card) and -1
' otherwise

asm             mov     ax,&h4F02               ;SVGA Set Mode
asm             mov     bx,&h0100               ;640x400x256 B
asm             int     &h10                    ;BIOS video

asm             cmp     ax,&h4F                 ;Did it work?
asm             je      SetupCont               ;Jump if so

asm             mov     ax,&h4F02               ;SVGA Set Mode
asm             mov     bx,&h0101               ;640x480x256 B
asm             int     &h10                    ;BIOS video

asm             cmp     ax,&h4F                 ;Did it work?
asm             je      SetupCont               ;if not

                FUNCTION = 0                    'function failed
                EXIT FUNCTION                   'return

    SetupCont:

asm             mov     dx,&h03C4               ;Sequencer Port
asm             mov     ax,&h0100               ;Synchronous reset
asm             out     dx,ax                   ;Send command
asm             mov     ax,&h0604               ;Disable Chain-4 Mode
asm             out     dx,ax                   ;Send command
asm             mov     ax,&h0F02               ;Enable all planes
asm             out     dx,ax                   ;Send command

asm             mov     dx,&h03CC               ;Misc Output Read Port
asm             in      al,dx                   ;Read byte
asm             and     al,&h07F                ;Set 400 scanlines
asm             or      al,&h040
asm             mov     dx,&h03C2               ;Misc Output Write Port
asm             out     dx,al                   ;Send command

asm             mov     dx,&h03D4               ;CRT Controller Port
asm             mov     cx,11                   ;11 words
asm             xor     bx,bx

    IG_outstr:
asm             mov     ax,cs:CRTCData[bx]
asm             out     dx,ax                   ;Send data
asm             inc     bx
asm             inc     bx
asm             loop    IG_outstr

asm             mov     dx,&h03C4               ;Sequencer Port
asm             mov     ax,&h0300               ;Restart Sequencer
asm             out     dx,ax                   ;Send command

asm             mov     ax,&h1130               ;Get pointer to
asm             mov     bh,3                    ;ROM 8x8 font
asm		push    bp
asm             int     &h10
asm		mov     bx,bp
asm		pop	bp
asm             mov     cs:FontPtr[0],bx        ;Save this pointer
asm             mov     cs:FontPtr[2],es

asm             push    di                      ;save di
asm		db 	%pushw
asm             dw      &hA000
asm             pop     es                      ;ES = video memory
asm             xor     di,di                   ;Zero DI, AX
asm             xor     ax,ax
asm             mov     cx,&h8000               ;Clear the screen
asm             rep     stosw
asm             pop     di                      ;restore di

asm             mov     word ptr cs:Colour,15   ;Set color to white
asm             mov     word ptr cs:MinX,0      ;Set UL corner to (0,0)
asm             mov     word ptr cs:MinY,0
asm             mov     word ptr cs:MaxX,639    ;Set LR corner to (639,399)
asm             mov     word ptr cs:MaxY,399
asm             mov     byte ptr cs:FontWidth,8
asm             mov     byte ptr cs:FontHeight,8

                FUNCTION = -1                   'function succesful
END FUNCTION

'**************************** CloseGraph

SUB CloseGraph () PUBLIC

asm             mov     ax, 3
asm             int     &h10

END SUB

'**************************** ClearScreen -- clears the screen

sub ClearScreen () public

asm		db 	%pushw
asm             dw      &hA000
asm             pop     es                      ;ES = video memory

asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     ax,&h0F02               ;Set Map Mask
asm             out     dx,ax

asm             push    di                      ;Save DI
asm             xor     ax,ax                   ;Clear AX
asm             xor     di,di                   ;Clear DI
asm             mov     cx,32000                ;CX = 32000 words
asm             rep     stosw                   ;Clear memory
asm             pop     di                      ;Restore di

end sub

'**************************** ClearViewport -- clears the viewport

sub ClearViewPort () public

asm             db      %pusha

asm		db 	%pushw
asm             dw      &hA000
asm             pop     es                      ;ES = video memory
asm             mov     ax,cs:MinX              ;Get corners of viewport
asm             mov     bx,cs:MinY              ;in AX, BX, CX, DX
asm             mov     cx,cs:MaxX
asm             mov     dx,cs:MaxY

asm             sub     dx,bx                   ;DX = height
asm             mov     di,ax                   ;Put X1, X2 in DI, SI
asm             mov     si,cx
asm             dw      %shr_di                 ;Divide them by 4
asm             db      2
asm             dw      %shr_si
asm             db      2
asm             push    ax                      ;Save AX
asm             mov     al,160                  ;Multiply Y by 160
asm             mul     bl
asm             add     di,ax                   ;Add to X1
asm             add     si,ax                   ;Add to X2
asm             pop     ax                      ;Restore AX

asm             and     ax,3                    ;AX = X1 mod 4
asm             and     cx,3                    ;CX = Y1 mod 4
asm             mov     bx,&h0101               ;BL = 1, BH = 1
asm             inc     cl                      ;Adjust shift count
asm             shl     bh,cl                   ;Calculate map mask
asm             dec     bh                      ;for the right side
asm             xchg    ax,cx                   ;Switch AX, CX
asm             shl     bl,cl                   ;for the left side
asm             neg     bl                      ;which is a little
asm             add     bl,16                   ;more complicated

asm             cmp     si,di                   ;Thin rectangle?
asm             je      cv_thin                 ;Jump if yes

    cv_loop:
asm             db      %pusha                  ;Save registers
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Left side
asm             out     dx,ax                   ;Send command
asm             xor     al,al                   ;AL = 0 (black)
asm             stosb                           ;Store byte
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bh                   ;Right side
asm             out     dx,ax                   ;Send command
asm             mov     byte ptr es:[si],0      ;Store byte

asm             mov     ax,&h0F02               ;Enable all planes
asm             out     dx,ax                   ;Send command
asm             mov     al,ch                   ;AL = color
asm             mov     ah,al                   ;AH = color
asm             mov     cx,si                   ;CX = distance
asm             sub     cx,di
asm             xor     dx,dx                   ;Clear DX
asm             shr     cx,1                    ;CX = words
asm             adc     dx,0                    ;DX = odd byte
asm             rep     stosw                   ;Store by words
asm             mov     cx,dx                   ;Get odd byte
asm             rep     stosb                   ;Store odd byte

asm             db      %popa                   ;Restore registers

asm             add     di,160                  ;Next line
asm             add     si,160
asm             dec     dx                      ;Dec line counter
asm             jns     cv_loop                 ;Loop back
asm             jmp     cv_done                 ;Return

    cv_thin:
asm             and     bl,bh                   ;AND masks together
asm             push    dx                      ;Save DX
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Short line
asm             out     dx,ax                   ;Send command
asm             pop     dx                      ;Restore DX
asm             xor     al,al                   ;AL = 0 (black)

    cvt_loop:
asm             mov     es:[di],al              ;Store byte
asm             add     di,160                  ;Next line
asm             dec     dx                      ;Dec line counter
asm             jns     cvt_loop                ;Loop back

    cv_done:
asm             db      %popa

end sub

'**************************** Dissolve - fancy clear screen

sub Dissolve () public

asm             db      %pusha                  ;Save registers
asm             push    ds
asm		db 	%pushw
asm             dw      &hA000
asm             pop     ds                      ;DS = video memory

asm             mov     si,&hFFFF               ;CH:SI = 03FFFFh
asm             mov     ch,3
asm             mov     dx,&h03C4               ;DX = Sequencer port

asm             mov     ax,&h0102               ;Set pixel(0, 0) to 0
asm             out     dx,ax
asm             xor     bx,bx
asm             mov     byte ptr [bx],0

asm             inc     bx                      ;CL:BX = 1
asm             xor     cl,cl

    di_loop:
asm             mov     ah,1                    ;Clear pixel CL:BX
asm             shl     ah,cl
asm             out     dx,ax
asm             mov     byte ptr [bx],0
asm             shr     cl,1                    ;shr cl:bx,1
asm             rcr     bx,1
asm             jnc     di_noXOR                ;jnc di_noXOR
asm             xor     bx,&h0400               ;xor cl:bx,020400h
asm             xor     cl,&h02
    di_noXOR:
asm             sub     si,1                    ;dec ch:si
asm             sbb     ch,0
asm             test    si,si                   ;jnz di_loop
asm             jne     di_loop
asm             test    ch,ch
asm             jne     di_loop

asm             pop     ds                      ;Restore registers
asm             db      %popa

end sub

'**************************** Ellipse -- draws an ellipse

sub Ellipse (byval X  as word, byval Y  as word, _
             byval XR as word, byval YR as word) public

                'Supply (X, Y) = center, XR = X radius, YR = Y radius

asm             mov     cx, X
asm             mov     dx, Y
asm             mov     ax, XR
asm             mov     bx, YR

asm             db      %pusha                  ;Save all registers
asm             push    ds
asm             push    cs                      ;DS = CS
asm             pop     ds

asm             mov     e_xnum,ax               ;Save xrad
asm             mov     e_ynum,bx               ;Save yrad
asm             cmp     ax,bx                   ;Radius = maxrad
asm             jg      el_l1
asm             mov     ax,bx
    el_l1:
asm             mov     e_denom,ax              ;Circle radius
asm             mov     si,cx                   ;SI = Center X
asm             mov     di,dx                   ;DI = Center Y
asm             xor     cx,cx                   ;X = 0
asm             mov     dx,ax                   ;Y = radius
asm             mov     xsq[0],cx               ;X = 0
asm             mov     xsq[2],cx
asm             push    dx                      ;Save DX
asm             mul     ax                      ;Square radius
asm             mov     ysq[0],ax               ;Y = r
asm             mov     ysq[2],dx
asm             mov     rsq[0],ax               ;r = r (obviously)
asm             mov     rsq[2],dx
asm             pop     dx                      ;Restore DX
asm             mov     al,Colour               ;AL = color

    el_loop:
asm             cmp     cx,dx                   ;X > Y?
asm             jle     el_loopn                ;If so, done
                GOTO    el_done
    el_loopn:
asm                                             ;Set 8 pixels, symmetric
asm             call    near ptr el_SetPix      ;setpixel cx+x, cy+y
asm             neg     dx
asm             call    near ptr el_SetPix      ;setpixel cx+x, cy-y
asm             neg     cx
asm             call    near ptr el_SetPix      ;setpixel cx-x, cy-y
asm             neg     dx
asm             call    near ptr el_SetPix      ;setpixel cx-x, cy+y
asm             xchg    cx,dx
asm             call    near ptr el_SetPix      ;setpixel cx+y, cy-x
asm             neg     cx
asm             call    near ptr el_SetPix      ;setpixel cx-y, cy-x
asm             neg     dx
asm             call    near ptr el_SetPix      ;setpixel cx-y, cy+x
asm             neg     cx
asm             call    near ptr el_SetPix      ;setpixel cx+y, cy+x
asm             xchg    cx,dx

asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0
asm             inc     cx                      ;X = X + 1
asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0

asm		push	ax
asm             mov     bx,xsq[0]               ;BX:BP = X
asm             mov     ax,xsq[2]
asm             add     bx,ysq[0]               ;BX:BP = X + Y
asm             adc     ax,ysq[2]
asm             sub     bx,rsq[0]               ;BX:BP = X + Y - R
asm             sbb     ax,rsq[2]

asm             test    ax,ax                   ;Equivalent to:
asm		pop	ax
asm             jg      el_DecY                 ;cmp BX:BP,DX
asm             jl      el_loop                 ;jle ci_loop
asm             cmp     bx,dx
asm             jle     el_loop

    el_DecY:
asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             dec     dx                      ;Y = Y - 1
asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             jmp     el_loop                 ;Loop back

    el_done:
asm             pop     ds                      ;Restore registers
asm             db      %popa

                EXIT SUB

el_SetPix:

asm             db      %pusha                  ;Save all registers
asm             push    ax
asm             mov     bx,e_denom              ;Central set pixel
asm             push    dx              ;setpixel(si + (cx * xr / (xr + yr)),
asm             mov     ax,e_xnum       ;         di + (dx * yr / (xr + yr)))
asm             imul    cx
asm             idiv    bx
asm             mov     cx,ax
asm             pop     dx
asm             mov     ax,e_ynum
asm             imul    dx
asm             idiv    bx
asm             mov     dx,ax
asm             add     cx,si
asm             add     dx,di
asm             pop     ax
asm             call    near ptr nSetPixel
asm             db      %popa                   ;Restore all registers
asm             retn

end sub

'**************************** FillArea -- Area fill from seed pixel.

sub FillArea (byval X as word, byval Y as word, _
              byval Colour as word) public

asm             mov     cx, X
asm             mov     dx, Y
asm             mov     al, byte ptr Colour

asm             ;Supply CX = x, DX = y, AL = color

asm             db      %pusha                  ;Save all registers
asm             push    ds                      ;DS = CS, ES = 0A000h
asm             push    es
asm             push    cs
asm             pop     ds
asm		db 	%pushw
asm             dw      &hA000
asm             pop     es                      ;ES = video memory

asm             xor     ah,ah
asm             mov     new_color,ax            ;Save new color
asm             call    near ptr nGetPixel
asm             cmp     ax,-1                   ;Out of bounds, done
asm             je      fa_2done
asm             cmp     ax,new_color            ;Same color, done
asm             je      fa_2done
asm             mov     old_color,ax            ;Save old color
asm             mov     stack_top,sp

asm             push    cx
asm             push    dx
asm             push    dx
asm             db      %pushb
asm             db      1
asm             inc     cx
asm             push    cx                      ;Push X+1 Y Y -1 (seed value)
asm             push    dx
asm             push    dx
asm             db      %pushb
asm             db      &hFF
asm             dec     cx                      ;Restore X
asm             jmp	short fa_loop
    fa_2done:
asm             jmp     fa_done

    fa_loop:
asm             cmp     sp,stack_top            ;While (stack not empty)
asm             je      fa_2done

asm             pop     bp                      ;Pop X Y1 Y2 dX
asm             pop     di
asm             pop     si
asm             pop     cx
asm             add     cx,bp                   ;X = X + dX

asm             push    ax                      ;Save AX, CX, DX
asm             push    cx
asm             push    dx
asm             push    cx
asm             push    dx
asm             and     cl,3                    ;CL = plane number
asm             mov     ah,cl                   ;Set read plane
asm             mov     dx,&h03CE
asm             mov     al,4
asm             out     dx,ax
asm             mov     ax,&h0102               ;Set write plane
asm             shl     ah,cl
asm             mov     dx,&h03C4
asm             out     dx,ax
asm             pop     dx                      ;Restore registers
asm             pop     cx
asm             pop     dx

asm             mov     dx,si                   ;Y = Y1

    fa_1_loop:
asm             cmp     dx,MinY                 ;Past left edge?
asm             jl      fa_1_done

'****************** GETPIXEL -- inlined

asm             cmp     cx,MinX                 ;X < MinX?
asm             jl      fa_gpbad1
asm             cmp     cx,MaxX                 ;X > MaxX?
asm             jg      fa_gpbad1

asm             push    bx                      ;Save registers
asm             push    dx
asm             mov     bx,cx                   ;DX = y, BX = CX = x
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             mov     ax,dx                   ;Multiply Y by 160
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     al,es:[bx]              ;Get byte
asm             mov     bl,es:[bx]
asm             or      al,bl
asm             xor     ah,ah
asm             pop     dx                      ;Restore registers
asm             pop     bx
asm             jmp     fa_return1              ;Return

    fa_gpbad1:
asm             mov     ax,-1                   ;Bad value, return -1

'****************** End GETPIXEL

    fa_return1:
asm             cmp     ax,old_color            ;Still in the area?
asm             jne     fa_1_done
asm             mov     ax,new_color            ;Set pixel with new color

'****************** SETPIXEL -- inlined

asm             db      %pusha                  ;Save registers
asm             mov     bx,cx                   ;DX = y, BX = CX = x
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             mov     ch,al                   ;Save color in CH
asm             mov     ax,dx                   ;Multiply Y by 160
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     es:[bx],ch              ;Put byte in memory
asm             db      %popa

'****************** End SETPIXEL

asm             dec     dx                      ;Decrement Y
asm             jmp     fa_1_loop               ;Loop back

    fa_1_done:
asm             cmp     dx,si                   ;if Y >= Y1, then skip it
asm		jl      fa_2
asm             jmp     fa_skipit

    fa_2:
asm             mov     bx,dx                   ;L = Y+1
asm             inc     bx

asm             cmp     bx,si                   ;if L < Y1,
asm             jge     fa_2_nope
asm             dec     si
asm             neg     bp
asm             push    cx                      ;push X L Y1-1 -dX
asm             push    bx
asm             push    si
asm             push    bp
asm             inc     si
asm             neg     bp

    fa_2_nope:
asm             mov     dx,si                   ;Y = Y1+1
asm             inc     si

    fa_3_loop:
asm             cmp     dx,MaxY                 ;Past right edge?
asm             jg      fa_4_done

'****************** GETPIXEL -- inlined

asm             cmp     cx,MinX                 ;X < MinX?
asm             jl      fa_gpbad2
asm             cmp     cx,MaxX                 ;X > MaxX?
asm             jg      fa_gpbad2

asm             push    bx                      ;Save registers
asm             push    dx
asm             mov     bx,cx                   ;DX = y, BX = CX = x
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             mov     ax,dx                   ;Multiply Y by 160
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     al,es:[bx]              ;Get byte
asm             mov     bl,es:[bx]
asm             or      al,bl
asm             xor     ah,ah
asm             pop     dx                      ;Restore registers
asm             pop     bx
asm             jmp     fa_return2              ;Return

    fa_gpbad2:
asm             mov     ax,-1                   ;Bad value, return -1

'****************** End GETPIXEL

    fa_return2:
asm             cmp     ax,old_color            ;Still in the area?
asm             jne     fa_4_done
asm             mov     ax,new_color            ;Set pixel with new color

'****************** SETPIXEL -- inlined

asm             db      %pusha                  ;Save registers
asm             mov     bx,cx                   ;DX = y, BX = CX = x
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             mov     ch,al                   ;Save color in CH
asm             mov     ax,dx                   ;Multiply Y by 160
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     es:[bx],ch              ;Put byte in memory
asm             db      %popa

'****************** End SETPIXEL

asm             inc     dx                      ;Increment Y
asm             jmp     fa_3_loop               ;Loop back

    fa_4_done:
asm             dec     dx
asm             push    cx                      ;Push X L Y-1 dX
asm             push    bx
asm             push    dx
asm             push    bp
asm             inc     dx

asm             inc     di
asm             cmp     dx,di                   ;If Y > Y2+1,
asm             jle     fa_5_nope
asm             dec     dx
asm             neg     bp
asm             push    cx                      ;push X Y2+1 Y-1 -dX
asm             push    di
asm             push    dx
asm             push    bp
asm             inc     dx
asm             neg     bp

    fa_5_nope:
asm             dec     di                      ;Restore Y2

    fa_skipit:
asm             inc     dx                      ;Increment Y

    fa_6_loop:
asm             cmp     dx,di                   ;Past Y2?
asm             jg      fa_6_done

'****************** GETPIXEL -- inlined

asm             cmp     cx,MinX                 ;X < MinX?
asm             jl      fa_gpbad3
asm             cmp     cx,MaxX                 ;X > MaxX?
asm             jg      fa_gpbad3

asm             push    bx                      ;Save registers
asm             push    dx
asm             mov     bx,cx                   ;DX = y, BX = CX = x
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             mov     ax,dx                   ;Multiply Y by 160
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     al,es:[bx]              ;Get byte
asm             mov     bl,es:[bx]
asm             or      al,bl
asm             xor     ah,ah
asm             pop     dx                      ;Restore registers
asm             pop     bx
asm             jmp     fa_return3              ;Return

    fa_gpbad3:
asm             mov     ax,-1                   ;Bad value, return -1

'****************** End GETPIXEL

    fa_return3:
asm             cmp     ax,old_color            ;Back in fill area?
asm             je      fa_6_done
asm             inc     dx                      ;Decrement Y
asm             jmp     fa_6_loop               ;Loop back

    fa_6_done:
asm             mov     bx,dx                   ;L = Y

asm             cmp     dx,di
asm             jg      fa_l1
asm             jmp     fa_3_loop               ;Loop back (3)
    fa_l1:
asm             jmp     fa_loop                 ;Loop back main

    fa_done:
asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa

end sub

'**************************** FillCircle -- draws a filled circle

sub FillCircle (byval X as word, byval Y as word, byval R as word) public

asm             mov     cx,X
asm             mov     dx,Y
asm             mov     ax,R

asm             ;Supply (CX, DX) = center, AX = radius

asm             db      %pusha                  ;Save all registers
asm             push    ds
asm             push    cs                      ;DS = CS
asm             pop     ds

asm             mov     si,cx                   ;SI = Center X
asm             mov     di,dx                   ;DI = Center Y
asm             xor     cx,cx                   ;X = 0
asm             mov     dx,ax                   ;Y = radius
asm             mov     xsq[0],cx               ;X = 0
asm             mov     xsq[2],cx
asm             push    dx                      ;Save DX
asm             mul     ax                      ;Square radius
asm             mov     ysq[0],ax               ;Y = r
asm             mov     ysq[2],dx
asm             mov     rsq[0],ax               ;r = r (obviously)
asm             mov     rsq[2],dx
asm             pop     dx                      ;Restore DX

    fc_loop:
asm             cmp     dx,0                    ;Y < 0?
asm             jl      fc_done                 ;If so, done

asm             mov     bp,xsq[0]               ;BX:BP = X
asm             mov     bx,xsq[2]
asm             add     bp,ysq[0]               ;BX:BP = X + Y
asm             adc     bx,ysq[2]
asm             sub     bp,rsq[0]               ;BX:BP = X + Y - R
asm             sbb     bx,rsq[2]

asm             test    bx,bx                   ;Equivalent to:
asm             jg      fc_DecY                 ;cmp BX:BP,DX
asm             jl      fc_IncX                 ;jl fc_IncX
asm             cmp     bp,dx
asm             jl      fc_IncX

    fc_DecY:
asm             mov     ax,si                   ;Line(CenX - X,
asm             mov     bx,di                   ;     CenY + Y,
asm             sub     ax,cx                   ;     CenX + X,
asm             add     bx,dx                   ;     CenY + Y)
asm             add     cx,si
asm             add     dx,di
asm             call    near ptr nLine
asm             sub     dx,di                   ;Line(CenX - X,
asm             neg     dx                      ;     CenY - Y,
asm             add     bx,dx                   ;     CenX + X,
asm             add     bx,dx                   ;     CenY - Y)
asm             add     dx,di
asm             call    near ptr nLine          ;Restore regs
asm             sub     cx,si
asm             sub     dx,di
asm             neg     dx

asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             dec     dx                      ;Y = Y - 1
asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             jmp     fc_loop                 ;Loop back

    fc_IncX:
asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0
asm             inc     cx                      ;X = X + 1
asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0
asm             jmp     fc_loop                 ;Loop back

    fc_done:
asm             pop     ds                      ;Restore registers
asm             db      %popa

end sub

'**************************** FillEllipse -- draws a filled ellipse

sub FillEllipse (byval X as word,  byval Y as word, _
                 byval XR as word, byval YR as word) public

asm             mov     cx,X
asm             mov     dx,Y
asm             mov     ax,XR
asm             mov     bx,YR

asm             ;Supply (CX, DX) = center, AX = X radius, BX = Y radius

asm             db      %pusha                  ;Save all registers
asm             push    ds
asm             push    cs                      ;DS = CS
asm             pop     ds

asm             mov     e_xnum,ax               ;Save xrad
asm             mov     e_ynum,bx               ;Save yrad
asm             cmp     ax,bx                   ;Radius = maxrad
asm             jg      fe_l1
asm             mov     ax,bx

    fe_l1:
asm             mov     e_denom,ax              ;Circle radius
asm             mov     si,cx                   ;SI = Center X
asm             mov     di,dx                   ;DI = Center Y
asm             xor     cx,cx                   ;X = 0
asm             mov     dx,ax                   ;Y = radius
asm             mov     xsq[0],cx               ;X = 0
asm             mov     xsq[2],cx
asm             push    dx                      ;Save DX
asm             mul     ax                      ;Square radius
asm             mov     ysq[0],ax               ;Y = r
asm             mov     ysq[2],dx
asm             mov     rsq[0],ax               ;r = r (obviously)
asm             mov     rsq[2],dx
asm             pop     dx                      ;Restore DX

    fe_loop:
asm             cmp     dx,0                    ;Y < 0?
asm		jge     fe_loop2		;If so,
asm		jmp	fe_done			;done
    fe_loop2:
asm             mov     bp,xsq[0]               ;BX:BP = X
asm             mov     bx,xsq[2]
asm             add     bp,ysq[0]               ;BX:BP = X + Y
asm             adc     bx,ysq[2]
asm             sub     bp,rsq[0]               ;BX:BP = X + Y - R
asm             sbb     bx,rsq[2]

asm             test    bx,bx                   ;Equivalent to:
asm             jg      fe_DecY                 ;cmp BX:BP,DX
asm             jl      fe_IncX                 ;jl fc_IncX
asm             cmp     bp,dx
asm             jl      fe_IncX

    fe_DecY:
asm             push    cx                      ;Save regs
asm             push    dx
asm             push    dx
asm             mov     bx,e_denom              ;cx = (cx * xr / (xr + yr)),
asm             mov     ax,e_xnum
asm             imul    cx
asm             idiv    bx
asm             mov     cx,ax

asm             pop     dx                      ;dx = (dx * yr / (xr + yr))
asm             mov     ax,e_ynum
asm             imul    dx
asm             idiv    bx
asm             mov     dx,ax

asm             mov     ax,si                   ;Line(CenX - X,
asm             mov     bx,di                   ;     CenY + Y,
asm             sub     ax,cx                   ;     CenX + X,
asm             add     bx,dx                   ;     CenY + Y)
asm             add     cx,si
asm             add     dx,di
asm             call    near ptr nLine

asm             sub     dx,di                   ;Line(CenX - X,
asm             neg     dx                      ;     CenY - Y,
asm             add     bx,dx                   ;     CenX + X,
asm             add     bx,dx                   ;     CenY - Y)
asm             add     dx,di
asm             call    near ptr nLine

asm             pop     dx                      ;Restore regs
asm             pop     cx

asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             dec     dx                      ;Y = Y - 1
asm             sub     xsq[0],dx               ;Y = Y - Y
asm             sbb     word ptr xsq[2],0
asm             jmp     fe_loop                 ;Loop back

    fe_IncX:
asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0
asm             inc     cx                      ;X = X + 1
asm             add     xsq[0],cx               ;X = X + X
asm             adc     word ptr xsq[2],0
asm             jmp     fe_loop                 ;Loop back

    fe_done:
asm             pop     ds                      ;Restore registers
asm             db      %popa

end sub

'**************************** FillRectangle -- draws a filled rectangle

sub FillRectAngle (byval X1 as word, byval Y1 as word, _
                           byval X2 as word, byval Y2 as word) public

asm             mov     ax,X1
asm             mov     bx,Y1
asm             mov     cx,X2
asm             mov     dx,Y2

asm             ;Supply AX=x1, BX=y1, CX=x2, DX=y2

asm             db      %pusha                  ;Push registers
asm             push    es
asm             db      %pushw
asm             dw      &hA000                  ;ES = video memory
asm             pop     es

asm             cmp     ax,cs:MaxX              ;X1 > MaxX?
asm             jg      fr_x1g
asm             cmp     ax,cs:MinX              ;X1 < MinX?
asm             jl      fr_x1l
    fr_cx2g:
asm             cmp     cx,cs:MaxX              ;X2 > MaxX?
asm             jg      fr_x2g
    fr_cx2l:
asm             cmp     cx,cs:MinX              ;X2 < MinX?
asm             jl      fr_x2l
    fr_checky:
asm             cmp     bx,cs:MaxY              ;Y1 > MaxY?
asm             jg      fr_y1g
    asm
asm             cmp     bx,cs:MinY              ;Y1 < MinY?
asm             jl      fr_y1l
    fr_cy2g:
asm             cmp     dx,cs:MaxY              ;Y2 > MaxY?
asm             jg      fr_y2g
    fr_cy2l:
asm             cmp     dx,cs:MinY              ;Y2 < MinY?
asm             jl      fr_y2l
asm             jmp     fr_cont                 ;Continue

    fr_x1g:
asm             cmp     cx,cs:MaxX              ;X1 > MaxX.
asm             jg      fr_done                 ;Both > MaxX?
asm             mov     ax,cs:MaxX              ;No, fix X1
asm             jmp     fr_cx2l

    fr_x1l:
asm             cmp     cx,cs:MinX              ;X1 < MinX.
asm             jl      fr_done                 ;Both < MinX?
asm             mov     ax,cs:MinX              ;No, fix X1
asm             jmp     fr_cx2g

    fr_x2g:
asm             mov     cx,cs:MaxX              ;X2 > MaxX, fix X2
asm             jmp     fr_checky

    fr_x2l:
asm             mov     cx,cs:MinX              ;X2 < MinX, fix X2
asm             jmp     fr_checky

    fr_y1g:
asm             cmp     dx,cs:MaxY              ;Y1 > MaxY.
asm             jg      fr_done                 ;Both > MaxY?
asm             mov     bx,cs:MaxY              ;No, fix Y1
asm             jmp     fr_cy2l

    fr_y1l:
asm             cmp     dx,cs:MinY              ;Y1 < MinY.
asm             jl      fr_done                 ;Both < MinY?
asm             mov     bx,cs:MinY              ;No, fix Y1
asm             jmp     fr_cy2g

    fr_y2g:
asm             mov     dx,cs:MaxY              ;Y2 > MaxY, fix Y2
asm             jmp     fr_cont

    fr_done:
asm             pop     es                      ;Restore registers
asm             db      %popa
                EXIT SUB

    fr_y2l:
asm             mov     dx,cs:MinY              ;Y2 < MinY, fix Y2

    fr_cont:
asm             mov     si,dx                   ;Get distances in SI, DI
asm             mov     di,cx
asm             sub     di,ax                   ;Is the X distance negative?
asm             jge     fr_xplus                ;Switch points and negate
asm             neg     di
asm             xchg    cx,ax

    fr_xplus:
asm             sub     si,bx                   ;Same thing for Y
asm             jge     fr_yplus
asm             neg     si
asm             xchg    dx,bx

    fr_yplus:
asm             sub     dx,bx                   ;DX = height
asm             mov     di,ax                   ;Put X1, X2 in DI, SI
asm             mov     si,cx
asm             dw      %shr_di                 ;Divide them by 4
asm             db      2
asm             dw      %shr_si
asm             db      2
asm             push    ax                      ;Save AX, DX
asm             push    dx
asm             mov     ax,160                  ;Multiply Y by 160
asm             mul     bx
asm             add     di,ax                   ;Add to X1
asm             add     si,ax                   ;Add to X2
asm             pop     dx                      ;Restore AX, DX
asm             pop     ax

asm             and     ax,3                    ;AX = X1 mod 4
asm             and     cx,3                    ;CX = Y1 mod 4
asm             mov     bx,&h0101               ;BL = 1, BH = 1
asm             inc     cl                      ;Adjust shift count
asm             shl     bh,cl                   ;Calculate map mask
asm             dec     bh                      ;for the right side
asm             xchg    ax,cx                   ;Switch AX, CX
asm             shl     bl,cl                   ;for the left side
asm             neg     bl                      ;which is a little
asm             add     bl,16                   ;more complicated

asm             cmp     si,di                   ;Thin rectangle?
asm             je      fr_thin                 ;Jump if yes

    fr_loop:
asm             db      %pusha                  ;Save all registers
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Left side
asm             out     dx,ax                   ;Send command
asm             mov     al,cs:Colour            ;AL = color
asm             mov     ch,al                   ;CH = color
asm             stosb                           ;Store byte
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bh                   ;Right side
asm             out     dx,ax                   ;Send command
asm             mov     es:[si],ch              ;Store byte

asm             mov     ax,&h0F02               ;Enable all planes
asm             out     dx,ax                   ;Send command
asm             mov     al,ch                   ;AL = color
asm             mov     ah,al                   ;AH = color
asm             mov     cx,si                   ;CX = distance
asm             sub     cx,di
asm             xor     dx,dx                   ;Clear DX
asm             shr     cx,1                    ;CX = words
asm             adc     dx,0                    ;DX = odd byte
asm             rep     stosw                   ;Store by words
asm             mov     cx,dx                   ;Get odd byte
asm             rep     stosb                   ;Store odd byte
asm             db      %popa                   ;Restore registers
asm             add     di,160                  ;Next line
asm             add     si,160
asm             dec     dx                      ;Dec line counter
asm             jns     fr_loop                 ;Loop back
asm             jmp     fr_done                 ;Return

    fr_thin:
asm             and     bl,bh                   ;AND masks together
asm             push    dx                      ;Save DX
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Short line
asm             out     dx,ax                   ;Send command
asm             pop     dx                      ;Restore DX
asm             mov     al,cs:Colour            ;AL = color

    frt_loop:
asm             mov     es:[di],al              ;Store byte
asm             add     di,160                  ;Next line
asm             dec     dx                      ;Dec line counter
asm             jns     frt_loop                ;Loop back
asm             jmp     fr_done

end sub

'**************************** GetImage -- gets a rectangular image

sub GetImage (byval X1 as word, byval Y1 as word, _
              byval X2 as word, byval Y2 as word, _
              byval Image AS DWORD) public

' Before calling GetImage een array Image has to be dimensioned
' with a size equal to (X2 - X1 + 1) * (Y2 - Y1 + 1) + 4 bytes.
' GetImage needs a far pointer to that array.

asm             db      %pusha                  ;Save registers
asm             push    ds
asm             push    es
asm             mov     ax,X1
asm             mov     bx,Y1
asm             mov     cx,X2
asm             mov     dx,Y2
asm             mov     si,word ptr Image[0]
asm             mov     ds,word ptr Image[2]

asm             ;Supply DS:SI=buffer, AX=x1, BX=y1, CX=x2, DX=y2

asm             push    ds                      ;ES:DI = DS:SI
asm             pop     es
asm             mov     di,si
asm             db      %pushw
asm             dw      &hA000                  ;DS = video memory
asm             pop     ds

asm             xchg    ax,cx                   ;CX = x, DX = y,
asm             xchg    bx,dx                   ;AX = width,
asm             sub     ax,cx                   ;BX = height
asm             sub     bx,dx
asm             inc     ax                      ;Adjust AX, BX
asm             inc     bx

asm             push    ax                      ;SI = offset
asm             mov     ax,dx                   ;(x + y * 160)
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             mov     bp,cx                   ;BP = horiz. position
asm             dw      %shr_cx
asm             db      2
asm             add     cx,ax
asm             pop     ax
asm             mov     si,cx

asm             stosw                           ;Store width
asm             mov     cx,ax                   ;CX = width
asm             mov     ax,bx                   ;AX = height
asm             stosw                           ;Store height

asm             mov     dx,&h03CE               ;Graphics Controller Port
asm             mov     ax,bp                   ;Horiz. position in AH
asm             mov     ah,al
asm             mov     al,4                    ;Set Read Plane command
asm             and     ah,3
asm             mov     cs:gi_plane,ah          ;gi_plane = starting plane
asm             mov     word ptr cs:gi_cplane,0 ;gi_cplane = 0

asm             db      %pusha                  ;Save all registers
asm             call    vertret                 ;Wait for vertical retrace
asm             db      %popa                   ;Restore registers

    gi_ploop:
asm             out     dx,ax                   ;Send command
asm             push    ax                      ;Save AX, BX, CX, SI
asm             push    bx
asm             push    cx
asm             push    si
asm             mov     bp,si                   ;BP = SI
asm             add     cx,3                    ;     (CX + 3 - Plane)
asm             sub     cx,cs:gi_cplane         ;CX = ----------------
asm             dw      %shr_cx                 ;            4
asm             db      2
asm             mov     cs:gi_width,cx          ;Save CX

    gi_loop:
asm             mov     cx,cs:gi_width          ;Get width
asm             shr     cx,1                    ;Divide by 2
asm             rep     movsw                   ;Move by words
asm             test    word ptr cs:gi_width,1  ;Odd byte?
asm             jz      gi_l1                   ;Jump if not
asm             movsb                           ;Move odd byte
    gi_l1:
asm             add     bp,160                  ;Next line
asm             mov     si,bp                   ;Offset in SI
asm             dec     bx                      ;Decrement counter
asm             jnz     gi_loop                 ;Loop back
asm             pop     si                      ;Restore registers
asm             pop     cx
asm             pop     bx
asm             pop     ax
asm             inc     ah                      ;Next plane
asm             cmp     ah,4                    ;Next byte too?
asm             jne     gi_l2                   ;Jump if not
asm             inc     si                      ;Increment offset
asm             xor     ah,ah                   ;Plane 0
    gi_l2:
asm             inc     word ptr cs:gi_cplane   ;Next plane
asm             cmp     ah,cs:gi_plane          ;Back to start?
asm             jne     gi_ploop                ;Jump if not

asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa
end sub

'**************************** GetPalette 

function GetPalette& (byval ColorNumber as word) public

' GetPalette& returns the palette numbered ColorNumber as
' 65536 * Red + 256 * Green + Blue

asm             cli
asm             call    vertret
asm             mov     dx,&h03C7
asm             mov     ax,ColorNumber          ;set palette register
asm             out     dx,al
asm             inc     dx
asm             inc     dx
asm             jmp     gpal_l1
    gpal_l1:
asm		jmp     gpal_l2
    gpal_l2:
asm		in      al,dx                   ;read red value
asm             mov     dx,ax
asm             jmp     gpal_l3
    gpal_l3:
asm		jmp     gpal_l4
    gpal_l4:
asm		in      al,dx                   ;read green
asm             mov     ah,al
asm             jmp     gpal_l5
    gpal_l5:
asm		jmp     gpal_l6
    gpal_l6:
asm		in      al,dx                   ;read blue
asm             sti

end function

'**************************** GetPixel -- reads a pixel

function GetPixel (byval X as word, byval Y as word) public

' GetPixel returns the color of the pixel at the given coordinates

asm             mov     cx, X
asm             mov     dx, Y
asm             call    near ptr nGetPixel

end function

'**************************** OutStr -- outputs a string

sub OutStr (byval X as word, byval Y as word, byval Text as string) public

                dim Temp as string
                dim TempPtr as byte ptr
                Temp = Text + chr$(0)
                TempPtr = strptr32(Temp)

asm             ;Supply DS:SI=string, CX=x, DX=y

asm             db      %pusha                  ;Save registers
asm             push    ds
asm             lds     si,TempPtr              ;let DS:SI point to string
asm             mov     cx,X
asm             mov     dx,Y
    os_loop:
asm             lodsb                           ;Load byte
asm             test    al,al                   ;Zero = done
asm             jz      os_done
asm             push    cx                      ;Save old X
asm             call    near ptr nOutChar       ;Output character
asm             pop     bx                      ;Restore old X
asm             cmp     cx,bx                   ;Did it work?
asm             jne     os_loop                 ;If yes, loop back
    os_done:
asm             pop     ds                      ;Restore registers
asm             db      %popa

end sub

'**************************** PutImage -- puts a rectangular image

sub PutImage (byval X as word, byval Y as word, _
              byval Image AS DWORD) public

' PutImage needs a far pointer to an array Image made by GetImage,
' containing the image to put on the screen

asm             db      %pusha                  ;Save registers
asm             push    ds
asm             push    es

asm             mov     cx,X
asm             mov     dx,Y
asm             mov     si,word ptr Image[0]
asm             mov     ds,word ptr Image[2]

asm             ;Supply DS:SI=buffer, CX=x, DX=y

asm             db      %pushw
asm             dw      &hA000                  ;ES = video memory
asm             pop     es

asm             lodsw                           ;Load width
asm             mov     bx,ax                   ;Put it in CX
asm             lodsw                           ;Load height

asm             cmp     cx,cs:MinX              ;X < MinX?
asm             jge     pi_do
asm             cmp     dx,cs:MinY              ;Y < MinY?
asm             jge     pi_do
asm             add     cx,bx                   ;Lower right corner
asm             add     dx,ax
asm             cmp     cx,cs:MaxX              ;X > MaxX?
asm             jle     pi_do
asm             cmp     dx,cs:MaxY              ;Y > MaxY?
asm             jle     pi_do
    pi_done:
asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa
                EXIT SUB

    pi_do:
asm             sub     cx,bx                   ;Back to upper left
asm             sub     dx,ax

asm             push    ax                      ;DI = offset
asm             mov     ax,dx                   ;(x + y * 160)
asm             dw      %shl_dx
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             mov     bp,cx                   ;BP = horiz. position
asm             dw      %shr_cx
asm             db      2
asm             add     cx,ax
asm             pop     ax
asm             mov     di,cx

asm             mov     cx,bx                   ;CX = width
asm             mov     bx,ax                   ;BX = height

asm             mov     dx,&h03C4               ;Sequencer Port
asm             mov     al,2                    ;Set Write Plane command
asm             mov     ah,&h11                 ;AH = 11h
asm             xchg    cx,bp                   ;Switch with CX
asm             and     cl,3                    ;CX = CX and 3
asm             shl     ah,cl                   ;AH = AH shl CX
asm             mov     cx,bp                   ;Restore CX
asm             mov     cs:pi_plane,ah          ;pi_plane = starting plane
asm             mov     word ptr cs:pi_cplane,0 ;pi_cplane = 0

    pi_ploop:
asm             out     dx,ax                   ;Send command
asm             push    ax                      ;Save AX, BX, CX, DI
asm             push    bx
asm             push    cx
asm             push    di
asm             mov     bp,di                   ;BP = DI
asm             add     cx,3                    ;     (CX + 3 - Plane)
asm             sub     cx,cs:pi_cplane         ;CX = ----------------
asm             dw      %shr_cx                 ;            4
asm             db      2
asm             mov     cs:pi_width,cx          ;Save CX

    pi_loop:
asm             mov     cx,cs:pi_width          ;Get width
asm             shr     cx,1                    ;Divide by 2
asm             rep     movsw                   ;Move by words
asm             test    word ptr cs:pi_width,1  ;Odd byte?
asm             jz      pi_l1                   ;Jump if not
asm             movsb                           ;Move odd byte
    pi_l1:
asm             add     bp,160                  ;Next line
asm             mov     di,bp                   ;Offset in SI
asm             dec     bx                      ;Decrement counter
asm             jnz     pi_loop                 ;Loop back

asm             pop     di                      ;Restore registers
asm             pop     cx
asm             pop     bx
asm             pop     ax
asm             shl     ah,1                    ;Next plane
asm             jnc     pi_l2                   ;Next byte too?
asm             inc     di                      ;Increment offset
asm             mov     ah,&h11                 ;Plane 0
    pi_l2:
asm             inc     word ptr cs:pi_cplane   ;Next plane
asm             cmp     ah,cs:pi_plane          ;Back to start?
asm             jne     pi_ploop                ;Jump if not
asm             jmp     pi_done

end sub

'**************************** PutCircle -- draws a circle

sub PutCircle (byval X as word, byval Y as word, _
               byval R as word) public

asm             mov     cx, X
asm             mov     dx, Y
asm             mov     ax, R

asm             ;Supply (CX, DX) = center, AX = radius

asm             db      %pusha                  ;Save all registers
asm             push    ds
asm             push    cs                      ;DS = CS
asm             pop     ds

asm             mov     si, cx                  ;SI = Center X
asm             mov     di, dx                  ;DI = Center Y
asm             xor     cx, cx                  ;X = 0
asm             mov     dx, ax                  ;Y = radius
asm             mov     xsq[0], cx              ;X = 0
asm             mov     xsq[2], cx
asm             push    dx                      ;Save DX
asm             mul     ax                      ;Square radius
asm             mov     ysq[0], ax              ;Y = r
asm             mov     ysq[2], dx
asm             mov     rsq[0], ax              ;r = r (obviously)
asm             mov     rsq[2], dx
asm             pop     dx                      ;Restore DX
asm             mov     al,Colour               ;AL = color

    ci_loop:
asm             cmp     cx, dx                  ;X > Y?
asm		jle	ci_loop2                ;If so,
asm		jmp	ci_done			;done
    ci_loop2:
asm                                             ;Set 8 pixels, symmetric
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx+x, cy+y)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     dx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx+x, cy-y)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     cx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx-x, cy-y)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     dx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx-x, cy+y)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             xchg    cx, dx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx+y, cy-x)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     cx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx-y, cy-x)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     dx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx-y, cy+x)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             neg     cx
asm             add     cx,si
asm             add     dx,di                   ;setpixel(cx+y, cy+x)
asm             call    near ptr nSetPixel
asm             sub     cx,si
asm             sub     dx,di
asm             xchg    cx, dx

asm             add     xsq[0], cx              ;X = X + X
asm             adc     word ptr xsq[2], 0
asm             inc     cx                      ;X = X + 1
asm             add     xsq[0], cx              ;X = X + X
asm             adc     word ptr xsq[2], 0

asm             mov     bp, xsq[0]               ;BX:BP = X
asm             mov     bx, xsq[2]
asm             add     bp, ysq[0]               ;BX:BP = X + Y
asm             adc     bx, ysq[2]
asm             sub     bp, rsq[0]               ;BX:BP = X + Y - R
asm             sbb     bx, rsq[2]

asm             test    bx, bx                  ;Equivalent to:
asm             jg      ci_DecY                 ;cmp BX:BP,DX
asm             jl      ci_loop3                ;jle ci_loop
asm             cmp     bp, dx
asm             jle     ci_loop3

    ci_DecY:
asm             sub     xsq[0], dx              ;Y = Y - Y
asm             sbb     word ptr xsq[2], 0
asm             dec     dx                      ;Y = Y - 1
asm             sub     xsq[0], dx              ;Y = Y - Y
asm             sbb     word ptr xsq[2], 0
    ci_loop3:
asm             jmp     ci_loop                 ;Loop back

    ci_done:
asm             pop     ds                      ;Restore registers
asm             db      %popa

end sub

'**************************** PutLine -- draws a line

sub PutLine (byval X1 as word, byval Y1 as word, _
             byval X2 as word, byval Y2 as word) public
asm             mov     ax,X1
asm             mov     bx,Y1
asm             mov     cx,X2
asm             mov     dx,Y2
asm             call    near ptr nLine
end sub

'**************************** Rectangle -- draws a rectangle

sub RectAngle (byval X1 as word, byval Y1 as word, _
               byval X2 as word, byval Y2 as word) public

asm             mov     ax,X1
asm             mov     bx,Y1
asm             mov     cx,X2
asm             mov     dx,Y2

asm             ;Supply AX=x1, BX=y1, CX=x2, DX=y2

asm             push    si                      ;Save SI
asm             mov     si,cx                   ;Save X2
asm             mov     cx,ax                   ;X2 = X1
asm             call    near ptr nLine          ;Draw top side
asm             mov     cx,si                   ;Restore X2
asm             mov     si,ax                   ;Save X1
asm             mov     ax,cx                   ;X1 = X2
asm             call    near ptr nLine          ;Draw bottom side
asm             mov     ax,si                   ;Restore X1
asm             mov     si,dx                   ;Save Y2
asm             mov     dx,bx                   ;Y2 = Y1
asm             call    near ptr nLine          ;Draw left side
asm             mov     dx,si                   ;Restore Y2
asm             mov     si,bx                   ;Save Y1
asm             mov     bx,dx                   ;Y1 = Y2
asm             call    near ptr nLine          ;Draw right side
asm             mov     bx,si                   ;Restore Y2
asm             pop     si                      ;Restore SI

end sub

'**************************** SetAllPalette

sub SetAllPalette (byval Palet AS DWORD) public

' SetAllPalette needs an long array Palet with 256 elements containing
' the palet to set. The palette values must be calculated as
' 65536 * Red + 256 * Green + Blue

asm             push    si
asm             push    ds
asm             cli
asm             call    vertret                 ;vertical retrace
asm             mov     dx,&h03C8
asm             xor     al,al                   ;start with color 0
asm             out     dx,al
asm             inc     dx
asm             mov     si,word ptr Palet[0]    ;DS:SI = pointer
asm             mov     ds,word ptr Palet[2]
asm             mov     cx,256                  ;256 colors
    spal_loop:
asm             lodsb                           ;load red value
asm             jmp     spal_l1                 ;delay
    spal_l1:
asm             jmp     spal_l2
    spal_l2:
asm             out     dx,al                   ;write it
asm             lodsb                           ;load green
asm             jmp     spal_l3                 ;delay
    spal_l3:
asm             jmp     spal_l4
    spal_l4:
asm             out     dx,al                   ;write it
asm             lodsb                           ;load blue
asm             jmp     spal_l5                 ;delay
    spal_l5:
asm             jmp     spal_l6
    spal_l6:
asm             out     dx,al                   ;write it
asm             jmp     spal_l7                 ;delay
    spal_l7:
asm             loop    spal_loop               ;loop back
asm             sti
asm             pop     ds
asm             pop     si
end sub

'**************************** SetColor

sub SetColor (byval Clr as byte) public
asm             mov     al,Clr
asm             mov     Colour,al
end sub

'**************************** SetFont

sub SetFont (byval Font AS DWORD) public

' SetFont needs a far pointer to a font array

asm             push    si
asm             push    ds
asm             lds     si,Font
asm             lodsw
asm             mov     cs:FontPtr[0],si
asm             mov     cs:FontPtr[2],ds
asm             mov     cs:FontWidth,al
asm             mov     cs:FontHeight,ah
asm             pop     ds
asm             pop     si

end sub

'**************************** SetFont8x8

sub SetFont8x8 () public
asm             push    bp
asm             mov     ax,&h1130
asm             mov     bh,3
asm             int     &h10
asm             mov     byte ptr cs:FontWidth,8
asm             mov     byte ptr cs:FontHeight,8
asm             mov     cs:FontPtr[0],bp
asm             mov     cs:FontPtr[2],es
asm             pop     bp
end sub

'**************************** SetFont8x16

sub SetFont8x16 () public
asm             push    bp
asm             mov     ax,&h1130
asm             mov     bh,6
asm             int     &h10
asm             mov     byte ptr cs:FontWidth,8
asm             mov     byte ptr cs:FontHeight,16
asm             mov     cs:FontPtr[0],bp
asm             mov     cs:FontPtr[2],es
asm             pop     bp
end sub

'**************************** SetPalette

sub SetPalette (byval ColorNumber as word, byval Value as long) public

asm             cli
asm             call    VertRet
asm             mov     dx,&h03C8
asm             mov     al,byte ptr ColorNumber ;set palette register
asm             out     dx,al
asm             inc     dx
asm             mov     al,byte ptr Value[2]    ;write red value
asm             jmp     spal_11
    spal_11:
asm             jmp     spal_12
    spal_12:
asm             out     dx,al
asm             mov     al,byte ptr Value[1]    ;write green
asm             jmp     spal_13
    spal_13:
asm             jmp     spal_14
    spal_14:
asm             out     dx,al
asm             mov     al,byte ptr Value[0]    ;write blue
asm             jmp     spal_15
    spal_15:
asm             jmp     spal_16
    spal_16:
asm             out     dx,al
asm             sti

end sub

'**************************** SetPixel - plots a pixel

sub SetPixel (byval X as word, byval Y as word, _
            byval Clr as byte) public

asm             mov     cx, X
asm             mov     dx, Y
asm             mov     al, Clr
asm             call    near ptr nSetPixel

end sub

'**************************** SetViewPort - sets the view port

sub SetViewPort (byval X1 as word, byval Y1 as word, _
                 byval X2 as word, byval Y2 as word) public

asm             mov     ax,X1
asm             mov     MinX,ax
asm             mov     ax,Y1
asm             mov     MinY,ax
asm             mov     ax,X2
asm             mov     MaxX,ax
asm             mov     ax,Y2
asm             mov     MaxY,ax

end sub

'**************************** VertRet - Waits for start of vertical retrace

sub VertRet () public

asm             mov     dx,&h03DA
    VRT1:
asm             in      al,dx
asm             test    al,8
asm             jnz     VRT1
    VRT2:
asm             in      al,dx
asm             test    al,8
asm             jz      VRT2

end sub

'**************************** nLine -- draws a line

nLine:

asm             ;Supply AX=x1, BX=y1, CX=x2, DX=y2

asm             db      %pusha                  ;Push registers
asm             push    ds
asm             push    es
asm             db      %pushw
asm             dw      &hA000                  ;ES = video memory
asm             pop     es
asm             push    cs                      ;DS = CS
asm             pop     ds
asm             cmp     bx,dx                   ;Is it horizontal?
asm             jne     l_checkv
asm             jmp     HorizLine
    l_checkv:
asm             cmp     ax,cx                   ;Vertical?
asm             jne     l_DiagLine
asm             jmp     VertLine

    l_DiagLine:
asm             cmp     ax,MaxX                 ;X1 > MaxX?
asm             jg      l_x1g
asm             cmp     ax,MinX                 ;X1 < MinX?
asm             jl      l_x1l
    l_cx2g:
asm             cmp     cx,MaxX                 ;X2 > MaxX?
asm             jg      l_x2g
    l_cx2l:
asm             cmp     cx,MinX                 ;X2 < MinX?
asm             jl      l_x2l
asm             jmp     l_checky

    l_x1g:
asm             cmp     cx,MaxX                 ;X1 > MaxX.
asm             jle     l_x1g2
asm             jmp     l_done                  ;Both > MaxX?
    l_x1g2:
asm             push    cx                      ;Save x2, y2
asm             push    dx
asm             sub     cx,ax
asm             sub     dx,bx                   ;Calculate the intersection:
asm             sub     ax,MaxX
asm             neg     ax                      ;     (MaxX - x1) * (y2 - y1)
asm             imul    dx                      ;y1 + -----------------------
asm             idiv    cx                      ;            (x2 - x1)
asm             add     bx,ax
asm             pop     dx                      ;Restore x2, y2
asm             pop     cx
asm             mov     ax,MaxX                 ;x1 = MaxX
asm             jmp     l_cx2l

    l_x1l:
asm             cmp     cx,MinX                 ;X1 < MinX.
asm             jge     l_x1l2
asm             jmp     l_done                  ;Both < MinX?
    l_x1l2:
asm             push    cx                      ;Save x2, y2
asm             push    dx
asm             sub     cx,ax
asm             sub     dx,bx                   ;Calculate the intersection:
asm             sub     ax,MinX
asm             neg     ax                      ;     (MinX - x1) * (y2 - y1)
asm             imul    dx                      ;y1 + -----------------------
asm             idiv    cx                      ;            (x2 - x1)
asm             add     bx,ax
asm             pop     dx                      ;Restore x2, y2
asm             pop     cx
asm             mov     ax,MinX                 ;x1 = MinX
asm             jmp     l_cx2g

    l_x2g:
asm             push    ax                      ;Save x1, y1
asm             push    bx
asm             sub     cx,ax
asm             sub     dx,bx                   ;Calculate the intersection:
asm             sub     ax,MaxX
asm             neg     ax                      ;     (MaxX - x1) * (y2 - y1)
asm             imul    dx                      ;y1 + -----------------------
asm             idiv    cx                      ;            (x2 - x1)
asm             add     bx,ax
asm             mov     dx,bx
asm             pop     bx                      ;Restore x1, y1
asm             pop     ax
asm             mov     cx,MaxX                 ;x2 = MaxX
asm             jmp     l_checky

    l_x2l:
asm             push    ax                      ;Save x1, y1
asm             push    bx
asm             sub     cx,ax
asm             sub     dx,bx                   ;Calculate the intersection:
asm             sub     ax,MinX
asm             neg     ax                      ;     (MinX - x1) * (y2 - y1)
asm             imul    dx                      ;y1 + -----------------------
asm             idiv    cx                      ;            (x2 - x1)
asm             add     bx,ax
asm             mov     dx,bx
asm             pop     bx                      ;Restore x1, y1
asm             pop     ax
asm             mov     cx,MinX                 ;x2 = MinX

    l_checky:
asm             cmp     bx,MaxY                 ;Y1 > MaxY?
asm             jg      l_y1g
asm             cmp     bx,MinY                 ;Y1 < MinY?
asm             jl      l_y1l
    l_cy2g:
asm             cmp     dx,MaxY                 ;Y2 > MaxY?
asm             jg      l_y2g
    l_cy2l:
asm             cmp     dx,MinY                 ;Y2 < MinY?
asm             jl      l_y2l
asm             jmp     l_cont1

    l_y1g:
asm             cmp     dx,MaxY                 ;Y1 > MaxY.
asm             jle     l_y1g2
asm             jmp     l_done                  ;Both > MaxY?
    l_y1g2:
asm             push    cx                      ;Save x2, y2
asm             push    dx
asm             xchg    ax,bx
asm             xchg    cx,dx
asm             sub     cx,ax                   ;Calculate the intersection:
asm             sub     dx,bx
asm             sub     ax,MaxY                 ;     (MaxY - y1) * (x2 - x1)
asm             neg     ax                      ;x1 + -----------------------
asm             imul    dx                      ;            (y2 - y1)
asm             idiv    cx
asm             add     ax,bx
asm             pop     dx                      ;Restore x2, y2
asm             pop     cx
asm             mov     bx,MaxY                 ;y1 = MaxX
asm             jmp     l_cy2l

    l_y1l:
asm             cmp     dx,MinY                 ;Y1 < MinY.
asm             jge     l_y1l2
asm             jmp     l_done                  ;Both < MinY?
    l_y1l2:
asm             push    cx                      ;Save x2, y2
asm             push    dx
asm             xchg    ax,bx
asm             xchg    cx,dx
asm             sub     cx,ax                   ;Calculate the intersection:
asm             sub     dx,bx
asm             sub     ax,MinY                 ;     (MinY - y1) * (x2 - x1)
asm             neg     ax                      ;x1 + -----------------------
asm             imul    dx                      ;            (y2 - y1)
asm             idiv    cx
asm             add     ax,bx
asm             pop     dx                      ;Restore x2, y2
asm             pop     cx
asm             mov     bx,MinY                 ;y1 = MinY
asm             jmp     l_cy2g

    l_y2g:
asm             push    ax                      ;Save x1, y1
asm             push    bx
asm             xchg    ax,bx
asm             xchg    cx,dx
asm             sub     cx,ax                   ;Calculate the intersection:
asm             sub     dx,bx
asm             sub     ax,MaxY                 ;     (MaxY - y1) * (x2 - x1)
asm             neg     ax                      ;x1 + -----------------------
asm             imul    dx                      ;            (y2 - y1)
asm             idiv    cx
asm             add     ax,bx
asm             mov     cx,ax
asm             pop     bx                      ;Restore x1, y1
asm             pop     ax
asm             mov     dx,MaxY                 ;y2 = MaxY
asm             jmp     l_cont1

    l_y2l:
asm             push    ax                      ;Save x1, y1
asm             push    bx
asm             xchg    ax,bx
asm             xchg    cx,dx
asm             sub     cx,ax                   ;Calculate the intersection:
asm             sub     dx,bx
asm             sub     ax,MinY                 ;     (MinY - y1) * (x2 - x1)
asm             neg     ax                      ;x1 + -----------------------
asm             imul    dx                      ;            (y2 - y1)
asm             idiv    cx
asm             add     ax,bx
asm             mov     cx,ax
asm             pop     bx                      ;Restore x1, y1
asm             pop     ax
asm             mov     dx,MinY                 ;y2 = MinY

    l_cont1:
asm             push    ax                      ;Save starting position
asm             push    bx
asm             sub     cx,ax                   ;get X and Y distance
asm             mov     word ptr xinc_str,1     ;Increments are 1
asm             mov     word ptr yinc_str,1
asm             jge     l_xplus                 ;X positive?
asm             neg     word ptr xinc_str       ;If not, negate
asm             neg     cx

    l_xplus:
asm             sub     dx,bx                   ;get Y distance
asm             jge     l_yplus                 ;Y positive?
asm             neg     word ptr yinc_str       ;If not, negate
asm             neg     dx

    l_yplus:
asm             mov     ax,xinc_str             ;Copy into the
asm             mov     xinc_diag,ax            ;diagonal values
asm             mov     ax,yinc_str
asm             mov     yinc_diag,ax
asm             cmp     cx,dx                   ;X < Y ?
asm             jle     l_ystr                  ;If so, go vertical
asm             mov     word ptr yinc_str,0     ;Otherwise go horizontal
asm             jmp     l_cont2                 ;At l_cont2: xinc, yinc_str are
    l_ystr:
asm             mov     word ptr xinc_str,0     ;the straight increments and xinc,
asm             xchg    cx,dx                   ;yinc_diag are the diag. increments
    l_cont2:
asm             mov     di,dx                   ;cx, dx are straight, perpendicular
asm             mov     deltav_str,di           ;Get dev. increments and start value
asm             sub     di,cx                   ;Str. incr = perp_dist
asm             mov     deltav_diag,di          ;Perp. incr = perp_dist - str_dist
asm             add     di,dx                   ;Starting value =
asm             sar     di,1                    ;(2 * perp_dist - str_dist) / 2
asm             mov     bp,di                   ;Put starting dev. in BP
asm             mov     di,cx                   ;Put length in DI
asm             inc     di                      ;One more point (ends)
asm             pop     si                      ;Get starting point in (cx,si)
asm             pop     cx
asm             mov     dx,&h03C4               ;DX = Sequencer port
asm             cmp     word ptr yinc_str,0     ;Zero on yinc_str?
asm             je      ldh_loop                ;If so, it's horizontal

    ldv_loop:
asm             mov     bx,cx                   ;SI = y, BX = CX = x
asm             mov     ax,si                   ;Multiply Y by 160
asm             dw      %shl_ax
asm             db      2
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             add     ax,si
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     al,Colour               ;CH = color
asm             mov     es:[bx],al              ;Store byte

asm             test    bp,bp                   ;Decide whether to go straight
asm             jle     ldv_gostr               ;or diagonal

    ldv_godiag:
asm             add     cx,xinc_diag            ;Go diagonal
asm             add     si,yinc_diag
asm             add     bp,deltav_diag

asm             mov     bx,cx                   ;BX = X and 3
asm             and     bx,3
asm             mov     al,2                    ;AL = setplane command
asm             mov     ah,Ptable[bx]           ;AH = plane value
asm             out     dx,ax                   ;Set plane

asm             dec     di                      ;Decrement counter
asm             jnz     ldv_loop                ;Loop back
asm             jmp     l_done

    ldv_gostr:
asm             add     si,yinc_str             ;Go straight
asm             add     bp,deltav_str
asm             dec     di                      ;Decrement counter
asm             jnz     ldv_loop                ;Loop back
asm             jmp     l_done

    ldh_loop:
asm             mov     scratch[0],cx           ;Save registers
asm             mov     scratch[2],si
asm             mov     bx,cx                   ;SI = y, BX = CX = x
asm             and     bx,3                    ;CL = bit plane
asm             mov     ah,Ptable[bx]           ;AH = plane mask
asm             mov     al,2                    ;AL = set plane command
asm             out     dx,ax                   ;Set plane
asm             mov     bx,cx
asm             mov     cx,si
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2
asm             dw      %shl_si                 ;Multiply Y by 160
asm             db      2
asm             add     cx,si
asm             dw      %shl_cx
asm             db      5
asm             add     bx,cx                   ;BX = offset
asm             mov     ch,Colour               ;CH = color
asm             mov     es:[bx],ch              ;Store byte
asm             mov     cx,scratch[0]           ;Restore registers
asm             mov     si,scratch[2]

asm             test    bp,bp                   ;Decide whether to go straight
asm             jg      ldh_godiag              ;or diagonal

    ldh_gostr:
asm             add     cx,xinc_str             ;Go straight
asm             add     bp,deltav_str
asm             dec     di                      ;Decrement counter
asm             jnz     ldh_loop                ;Loop back
asm             jmp     l_done

    ldh_godiag:
asm             add     cx,xinc_diag           ;Go diagonal
asm             add     si,yinc_diag
asm             add     bp,deltav_diag
asm             dec     di                      ;Decrement counter
asm             jnz     ldh_loop                ;Loop back
asm             jmp     l_done


    HorizLine:
asm             cmp     bx,MaxY                 ;Y > MaxY?
asm             jg      lh_done
asm             cmp     bx,MinY                 ;Y < MinY?
asm             jl      lh_done
asm             cmp     ax,MaxX                 ;X1 > MaxX?
asm             jg      lh_x1g
asm             cmp     ax,MinX                 ;X1 < MinX?
asm             jl      lh_x1l
    lh_cx2g:
asm             cmp     cx,MaxX                 ;X2 > MaxX?
asm             jg      lh_x2g
    lh_cx2l:
asm             cmp     cx,MinX                 ;X2 < MinX?
asm             jl      lh_x2l
asm             jmp     lh_cont                 ;Continue

    lh_x1g:
asm             cmp     cx,MaxX                 ;X1 > MaxX.
asm             jg      lh_done                 ;Both > MaxX?
asm             mov     ax,MaxX                 ;No, fix X1
asm             jmp     lh_cx2g

    lh_x1l:
asm             cmp     cx,MinX                 ;X1 < MinX.
asm             jl      lh_done                 ;Both < MinX?
asm             mov     ax,MinX                 ;No, fix X1
asm             jmp     lh_cx2g

    lh_x2g:
asm             mov     cx,MaxX                 ;X2 > MaxX, fix X2
asm             jmp     lh_cont

    lh_done:
asm		jmp     l_done

    lh_x2l:
asm             mov     cx,MinX                 ;X2 < MinX, fix X2
    lh_cont:
asm             cmp     cx,ax                   ;X1 > X2?
asm             jge     lh_xplus
asm             xchg    cx,ax                   ;Switch them

    lh_xplus:
asm             mov     di,ax                   ;Put X1, X2 in DI, SI
asm             mov     si,cx
asm             dw      %shr_di                 ;Divide them by 4
asm             db      2
asm             dw      %shr_si
asm             db      2
asm             push    ax                      ;Save AX, DX
asm             push    dx
asm             mov     ax,160                  ;Multiply Y by 160
asm             mul     bx
asm             add     di,ax                   ;Add to X1
asm             add     si,ax                   ;Add to X2
asm             pop     dx                      ;Restore AX, DX
asm             pop     ax

asm             and     ax,3                    ;AX = X1 mod 4
asm             and     cx,3                    ;CX = Y1 mod 4
asm             mov     bx,&h0101               ;BL = 1, BH = 1
asm             inc     cl                      ;Adjust shift count
asm             shl     bh,cl                   ;Calculate map mask
asm             dec     bh                      ;for the right side
asm             xchg    ax,cx                   ;Switch AX, CX
asm             shl     bl,cl                   ;for the left side
asm             neg     bl                      ;which is a little
asm             add     bl,16                   ;more complicated

asm             cmp     si,di                   ;Short line?
asm             je      lh_short                ;Jump if yes

asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Left side
asm             out     dx,ax                   ;Send command
asm             mov     al,Colour               ;AL = color
asm             mov     ch,al                   ;CH = color
asm             stosb                           ;Store byte
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bh                   ;Right side
asm             out     dx,ax                   ;Send command
asm             mov     es:[si],ch              ;Store byte

asm             mov     ax,&h0F02               ;Enable all planes
asm             out     dx,ax                   ;Send command
asm             mov     al,ch                   ;AL = color
asm             mov     ah,al                   ;AH = color
asm             mov     cx,si                   ;CX = distance
asm             sub     cx,di
asm             xor     dx,dx                   ;Clear DX
asm             shr     cx,1                    ;CX = words
asm             adc     dx,0                    ;DX = odd byte
asm             rep     stosw                   ;Store by words
asm             mov     cx,dx                   ;Get odd byte
asm             rep     stosb                   ;Store odd byte
asm             jmp     l_done                  ;Return

    lh_short:
asm             and     bl,bh                   ;AND masks together
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     al,2                    ;Set Map Mask
asm             mov     ah,bl                   ;Short line
asm             out     dx,ax                   ;Send command
asm             mov     al,Colour               ;AL = color
asm             mov     es:[di],al              ;Store byte
asm             jmp     l_done                  ;Return

    VertLine:
asm             cmp     ax,MaxX                 ;X > MaxX?
asm             jg      lv_done
asm             cmp     ax,MinX                 ;X < MinX?
asm             jl      lv_done
asm             cmp     bx,MaxY                 ;Y1 > MaxY?
asm             jg      lv_y1g
asm             cmp     bx,MinY                 ;Y1 < MinY?
asm             jl      lv_y1l
    lv_cy2g:
asm             cmp     dx,MaxY                 ;Y2 > MaxY?
asm             jg      lv_y2g
    lv_cy2l:
asm             cmp     dx,MinY                 ;Y2 < MinY?
asm             jl      lv_y2l
asm             jmp     lv_cont                 ;Continue

    lv_y1g:
asm             cmp     dx,MaxY                 ;Y1 > MaxY.
asm             jg      lv_done                 ;Both > MaxY?
asm             mov     bx,MaxY                 ;No, fix Y1
asm             jmp     lv_cy2g

    lv_y1l:
asm             cmp     dx,MinY                 ;Y1 < MinY.
asm             jl      lv_done                 ;Both < MinY?
asm             mov     bx,MinY                 ;No, fix Y1
asm             jmp     lv_cy2g

    lv_y2g:
asm             mov     dx,MaxY                 ;Y2 > MaxY, fix Y2
asm             jmp     lv_cont

    lv_done:
asm		jmp     l_done

    lv_y2l:
asm             mov     dx,MinY                 ;Y2 < MinY, fix Y2

    lv_cont:
asm             sub     dx,bx                   ;Get Y distance
asm             jge     lv_yplus                ;Negative?
asm             add     bx,dx                   ;BX = previous DX
asm             neg     dx                      ;Make it positive

    lv_yplus:
asm             push    ax                      ;Save X
asm             mov     ax,bx                   ;AX = Y * 160
asm             dw      %shl_bx
asm             db      2
asm             add     ax,bx
asm             dw      %shl_ax
asm             db      5
asm             pop     bx                      ;BX = X
asm             mov     cl,bl                   ;CL = low 2 bits of X
asm             and     cl,3
asm             dw      %shr_bx                 ;X = X / 4
asm             db      2
asm             add     ax,bx                   ;AX = X + Y * 160
asm             mov     di,ax                   ;Offset in DI
asm             mov     ah,1                    ;AH = 1
asm             shl     ah,cl                   ;AH = bit plane mask
asm             mov     al,2                    ;AL = 2: Set Map Mask
asm             push    dx                      ;Save DX
asm             mov     dx,&h03C4               ;Sequencer Port
asm             out     dx,ax                   ;Send command
asm             pop     dx                      ;Restore DX

asm             mov     al,Colour               ;AL = color
asm             push    es                      ;DS = ES
asm             pop     ds

    lv_loop:
asm             mov     [di],al                 ;Write pixel
asm             add     di,160                  ;Next line
asm             dec     dx                      ;Done?
asm             jns     lv_loop                 ;Loop back
asm             jmp     l_done                  ;Return

    l_done:
asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa
asm             retn                            ;Return

'**************************** OutChar -- outputs a character

nOutChar:
asm             ;Supply AL=char, CX=x, DX=y, returns CX=x+width

asm             db      %pusha                  ;Save registers
asm             push    ds
asm             push    es
asm		db 	%pushw
asm             dw      &hA000
asm             pop     es                      ;ES = video memory
asm             lds     si,FontPtr              ;DS:SI = font

asm             cmp     cx,MinX                 ;X < MinX?
asm             jl      oc_bad
asm             cmp     dx,MinY                 ;Y < MinY?
asm             jl      oc_bad
asm             add     cl,FontWidth            ;Lower right corner
asm             adc     ch,0
asm             add     dl,FontHeight
asm             adc     dh,0
asm             cmp     cx,MaxX                 ;X > MaxX?
asm             jg      oc_bad
asm             cmp     dx,MaxY                 ;Y > MaxY?
asm             jg      oc_bad
asm             sub     cl,FontWidth            ;Back to upper left
asm             sbb     ch,0
asm             sub     dl,FontHeight
asm             sbb     dh,0

asm             push    ax                      ;Save AX, CX
asm             push    cx
asm             and     cl,3                    ;Calculate map mask
asm             mov     ch,1
asm             shl     ch,cl
asm             mov     init_mask,ch            ;Save mask
asm             pop     cx                      ;Restore CX
asm             mov     ax,dx                   ;Get offset, which is
asm             dw      %shl_dx                 ;x / 4 + 160 * y
asm             db      2
asm             add     ax,dx
asm             dw      %shl_ax
asm             db      5
asm             dw      %shr_cx
asm             db      2
asm             add     cx,ax
asm             mov     di,cx                   ;Offset in DI
asm             pop     ax                      ;Restore AX
asm             jmp     oc_CheckFont

    oc_bad:                                     'Invalid height or width
asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa
asm		retn

    oc_CheckFont:
asm             cmp     byte ptr FontWidth,6    ;6 wide?
asm             jne     oc_8check
		goto    oc_6wide
    oc_8check:
asm             cmp     byte ptr FontWidth,8    ;8 wide?
asm             jne     oc_12check
		goto    oc_8wide
    oc_12check:
asm             cmp     byte ptr FontWidth,12   ;12 wide?
asm             je      oc_12wide
asm             cmp     byte ptr FontWidth,16   ;16 wide?
asm             jne     oc_bad

    oc_16wide:
asm             mov     cl,FontHeight           ;CX = font height
asm             xor     ch,ch
asm             mul     cl                      ;Point to specific char
asm             add     si,ax
asm             add     si,ax

    oc_16loop:
asm             mov     bp,[si]                 ;Load word
asm             mov     bx,&h8000
asm             add     si,2

asm             push    cx                      ;Save CX, DI
asm             push    di
asm             mov     ah,init_mask            ;Get initial mask
asm             mov     al,2                    ;Set Map Mask command
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     cx,16                   ;16 wide

    oc_16pxloop:
asm             test    bp,bx                   ;Is bit set?
asm             jz      oc_16px1                ;Jump if not
asm             out     dx,ax                   ;Set sequencer
asm             push    ax                      ;Save AX
asm             mov     al,Colour               ;Put color byte
asm             mov     es:[di],al              ;on screen
asm             pop     ax
    oc_16px1:
asm             shl     ah,1                    ;Next pixel
asm             shr     bx,1
asm             test    ah,16                   ;Byte boundary?
asm             jz      oc_16pxlb               ;Jump if not
asm             mov     ah,1                    ;Next byte,
asm             inc     di                      ;mask = 0001
    oc_16pxlb:
asm             dec     cx                      ;Loop back
asm             jnz     oc_16pxloop

asm             pop     di                      ;Restore CX, DI
asm             pop     cx
asm             add     di,160                  ;Next line
asm             dec     cx                      ;Loop back
asm             jnz     oc_16loop
                goto    oc_done

    oc_12wide:
asm             mov     cl,FontHeight           ;CX = font height
asm             xor     ch,ch
asm             mul     cl                      ;Point to specific char
asm             add     si,ax
asm             add     si,ax

    oc_12loop:
asm             mov     bp,[si]                 ;Load word
asm             mov     bx,&h8000
asm             add     si,2

asm             push    cx                      ;Save CX, DI
asm             push    di
asm             mov     ah,init_mask            ;Get initial mask
asm             mov     al,2                    ;Set Map Mask command
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     cx,12                   ;12 wide

    oc_12pxloop:
asm             test    bp,bx                   ;Is bit set?
asm             jz      oc_12px1                ;Jump if not
asm             out     dx,ax                   ;Set sequencer
asm             push    ax                      ;Save AX
asm             mov     al,Colour               ;Put color byte
asm             mov     es:[di],al              ;on screen
asm             pop     ax
    oc_12px1:
asm             shl     ah,1                    ;Next pixel
asm             shr     bx,1
asm             test    ah,16                   ;Byte boundary?
asm             jz      oc_12pxlb               ;Jump if not
asm             mov     ah,1                    ;Next byte,
asm             inc     di                      ;mask = 0001
    oc_12pxlb:
asm             dec     cx                      ;Loop back
asm             jnz     oc_12pxloop

asm             pop     di                      ;Restore CX, DI
asm             pop     cx
asm             add     di,160                  ;Next line
asm             dec     cx                      ;Loop back
asm             jnz     oc_12loop
                goto    oc_done

    oc_8wide:
asm             mov     cl,FontHeight           ;CX = font height
asm             xor     ch,ch
asm             mul     cl                      ;Point to specific char
asm             add     si,ax

    oc_8loop:
asm             mov     bl,[si]                 ;Load byte
asm             mov     bh,&h80
asm             inc     si

asm             push    cx                      ;Save CX, DI
asm             push    di
asm             mov     ah,init_mask            ;Get initial mask
asm             mov     al,2                    ;Set Map Mask command
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     cx,8                    ;8 wide

    oc_8pxloop:
asm             test    bl,bh                   ;Is bit set?
asm             jz      oc_8px1                 ;Jump if not
asm             out     dx,ax                   ;Set sequencer
asm             push    ax                      ;Save AX
asm             mov     al,Colour               ;Put color byte
asm             mov     es:[di],al              ;on screen
asm             pop     ax
    oc_8px1:
asm             shl     ah,1                    ;Next pixel
asm             shr     bh,1
asm             test    ah,16                   ;Byte boundary?
asm             jz      oc_8pxlb                ;Jump if not
asm             mov     ah,1                    ;Next byte,
asm             inc     di                      ;mask = 0001
    oc_8pxlb:
asm             dec     cx                      ;Loop back
asm             jnz     oc_8pxloop

asm             pop     di                      ;Restore CX, DI
asm             pop     cx
asm             add     di,160                  ;Next line
asm             dec     cx                      ;Loop back
asm             jnz     oc_8loop
                goto    oc_done

    oc_6wide:
asm             mov     cl,FontHeight           ;CX = font height
asm             xor     ch,ch
asm             mul     cl                      ;Point to specific char
asm             add     si,ax

    oc_6loop:
asm             mov     bl,[si]                 ;Load byte
asm             mov     bh,&h80
asm             inc     si

asm             push    cx                      ;Save CX, DI
asm             push    di
asm             mov     ah,init_mask            ;Get initial mask
asm             mov     al,2                    ;Set Map Mask command
asm             mov     dx,&h03C4               ;Sequencer port
asm             mov     cx,6                    ;6 wide

    oc_6pxloop:
asm             test    bl,bh                   ;Is bit set?
asm             jz      oc_6px1                 ;Jump if not
asm             out     dx,ax                   ;Set sequencer
asm             push    ax                      ;Save AX
asm             mov     al,Colour               ;Put color byte
asm             mov     es:[di],al              ;on screen
asm             pop     ax
    oc_6px1:
asm             shl     ah,1                    ;Next pixel
asm             shr     bh,1
asm             test    ah,16                   ;Byte boundary?
asm             jz      oc_6pxlb                ;Jump if not
asm             mov     ah,1                    ;Next byte,
asm             inc     di                      ;mask = 0001
    oc_6pxlb:
asm             dec     cx                      ;Loop back
asm             jnz     oc_6pxloop

asm             pop     di                      ;Restore CX, DI
asm             pop     cx
asm             add     di,160                  ;Next line
asm             dec     cx                      ;Loop back
asm             jnz     oc_6loop

    oc_done:
asm             pop     es                      ;Restore registers
asm             pop     ds
asm             db      %popa
asm             add     cl,FontWidth            ;Advance X
asm             adc     ch,0
asm		retn

'**************************** nGetPixel -- reads a pixel

nGetPixel:
asm             ;Supply CX=x, DX=y, returns AL=color

asm             cmp     cx,cs:MinX              ;X < MinX?
asm             jl      gp_bad
asm             cmp     dx,cs:MinY              ;Y < MinY?
asm             jl      gp_bad
asm             cmp     cx,cs:MaxX              ;X > MaxX?
asm             jg      gp_bad
asm             cmp     dx,cs:MaxY              ;Y > MaxY?
asm             jg      gp_bad

asm             push    bx                      ;Save registers
asm             push    dx
asm             push    si
asm             push    ds
asm		db 	%pushw
asm             dw      &hA000
asm             pop     ds                      ;DS = video memory

asm             mov     si,dx                   ;SI = y, BX = CX = x
asm             mov     bx,cx
asm             mov     ah,bl
asm             and     ah,3                    ;AH = bit plane
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2

asm             mov     dx,&h03CE               ;Graphics port
asm             mov     al,4                    ;Set Read Plane
asm             out     dx,ax

asm             mov     ax,si                   ;Multiply Y by 160
asm             dw      %shl_si
asm             db      2
asm             add     ax,si
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     al,[bx]                 ;Get byte
asm             mov     bl,[bx]
asm             or      al,bl
asm             xor     ah,ah

asm             pop     ds                      ;Restore registers
asm             pop     si
asm             pop     dx
asm             pop     bx
asm             retn                            ;Return

    gp_bad:
asm             mov     ax,-1                   ;Bad value, return -1
    SetPage:
asm             retn                            ;Dummy for set-page

'**************************** nSetPixel -- plots a pixel

nSetPixel:

asm             ;Supply CX=x, DX=y, AL=color

asm             cmp     cx,cs:MinX              ;X < MinX?
asm             jl      sp_done
asm             cmp     dx,cs:MinY              ;Y < MinY?
asm             jl      sp_done
asm             cmp     cx,cs:MaxX              ;X > MaxX?
asm             jg      sp_done
asm             cmp     dx,cs:MaxY              ;Y > MaxY?
asm             jg      sp_done

asm             db      %pusha                  ;Save registers
asm             push    ds
asm		db 	%pushw
asm             dw      &hA000
asm             pop     ds                      ;DS = video memory

asm             mov     si,dx                   ;SI = y, BX = CX = x
asm             mov     bx,cx
asm             and     cl,3                    ;BX = bit plane
asm             mov     ch,al                   ;Save color in CH
asm             mov     ax,&h0102               ;AH = plane mask, AL = command
asm             shl     ah,cl
asm             dw      %shr_bx                 ;BX = X offset
asm             db      2

asm             mov     dx,&h03C4               ;Sequencer port
asm             out     dx,ax

asm             mov     ax,si                   ;Multiply Y by 160
asm             dw      %shl_si
asm             db      2
asm             add     ax,si
asm             dw      %shl_ax
asm             db      5
asm             add     bx,ax                   ;BX = offset
asm             mov     [bx],ch                 ;Put byte in memory

asm             pop     ds                      ;Restore registers
asm             db      %popa

    sp_done:
asm             retn                            ;Return


