;****************** AGIMOUSE.ASM -- Mouse module for AGI

.model large, basic

.186

.code

MouseShown      db      0
MouseX          dw      0
MouseY          dw      0

;**************************** showmouse -- Turn mouse cursor on

ShowMouse       Proc    Far

                pushf
                cli
                mov     cs:[MouseShown],1
                call    near ptr PutCursor
                popf
                ret

ShowMouse       EndP

;**************************** hidemouse -- Turn mouse cursor off

HideMouse       Proc    Far

                pushf
                cli
                mov     cs:[MouseShown],0
                call    near ptr PutScrn
                popf
                ret

HideMouse       EndP

;**************************** getmouse -- Get mouse position

GetMouse        Proc    X:ptr word, Y:ptr word

                mov     ax,3
                int     33h
                mov     ax,bx

                cmp     word ptr X[2], 0
                je      gm_NoX
                les     bx, X
                mov     es:[bx], cx

gm_NoX:         cmp     word ptr Y[2], 0
                je      gm_NoY
                les     bx, Y
                mov     es:[bx], dx

gm_NoY:         ret

GetMouse        EndP

;**************************** setmouse -- Set mouse position

SetMouse        Proc    X:word, Y:word

                mov     ax, 4
                mov     cx, X
                mov     dx, Y
                int     33h

                ret

SetMouse        EndP

;**************************** getmousepress -- Get mouse press info

GetMousePress   Proc    Button:word, X:ptr word, Y:ptr word

                mov     ax, 5
                mov     bx, Button
                int     33h
                mov     ax, bx

                cmp     word ptr X[2], 0
                je      gp_NoX
                les     bx, X
                mov     es:[bx], cx

gp_NoX:         cmp     word ptr Y[2], 0
                je      gp_NoY
                les     bx, Y
                mov     es:[bx], dx

gp_NoY:         ret

GetMousePress   EndP

;**************************** getmouserelease -- Get mouse release info

GetMouseRelease Proc    Button:word, X:ptr word, Y:ptr word

                mov     ax,6
                mov     bx,Button
                int     33h
                mov     ax,bx

                cmp     word ptr X[2],0
                je      gr_NoX
                les     bx,X
                mov     es:[bx],cx

gr_NoX:         cmp     word ptr Y[2],0
                je      gr_NoY
                les     bx,Y
                mov     es:[bx],dx

gr_NoY:         ret

GetMouseRelease EndP

;**************************** limitmouse -- Define mouse limits

LimitMouse      Proc    MinX:word, MinY:word, MaxX:word, MaxY:word

                pushf
                cli

                mov     ax,7
                mov     cx,MinX
                mov     dx,MaxX
                int     33h
                mov     ax,8
                mov     cx,MinY
                mov     dx,MaxY
                int     33h

                popf
                ret

LimitMouse      EndP

;**************************** initmouse -- Initialize AGI mouse driver

InitMouse       Proc    Far

                push    ds
                push    di
                mov     ax,40h                  ;DS = 40h
                mov     ds,ax
                mov     dl,12h                  ;Fake mode 12h
                mov     di,49h
                xchg    dl,[di]

                xor     ax,ax                   ;Initialize mouse
                int     33h                     ;Mouse interrupt
                xchg    dl,[di]                 ;Fix BIOS mode
                pop     di                      ;Restore DI
                pop     ds                      ;Restore DS
                mov     cs:[MouseShown],0       ;Mouse is hidden

                push    ax                      ;Save AX for return value
                mov     ax,8                    ;Set Y limits
                xor     cx,cx                   ;(0 - 398)
                mov     dx,398
                int     33h                     ;Mouse interrupt

                mov     ax,4                    ;Set mouse position
                mov     cx,320                  ;(320, 200)
                mov     dx,200
                mov     cs:[MouseX],cx          ;Put in variables
                mov     cs:[MouseY],dx
                int     33h                     ;Mouse interrupt

                mov     ax,0Fh                  ;Set Mickey/Pixel Ratio
                mov     cx,8                    ;8 horizontal, 10 vertical
                mov     dx,10                   ;Square to within 4%
                int     33h                     ;Mouse Interrupt

                mov     ax,0Ch                  ;Set Mouse Handler
                mov     cx,1                    ;for motion event
                push    es
                push    cs                      ;ES = CS
                pop     es
                mov     dx,offset _MouseMove_   ;DX = offset of handler
                int     33h                     ;Mouse interrupt
                pop     es

                pop     ax
                ret                             ;Return

InitMouse       EndP

;**************************** endmouse -- De-activate AGI mouse driver

EndMouse        Proc    Far

                xor     ax,ax                   ;Re-initialize standard driver
                int     33h
                ret

EndMouse        EndP

;**************************** MouseMove -- Mouse motion handler

_MouseMove_     Proc    Far

                pushf                           ;Save flags
                cli                             ;No interrupts
                pusha                           ;Save registers
                cmp     cs:[MouseShown],0
                je      @F                      ;If mouse is shown,
                call    near ptr PutScrn        ;restore screen area
@@:             mov     cs:[MouseX],cx
                mov     cs:[MouseY],dx          ;Move mouse pointer
                cmp     cs:[MouseShown],0
                je      @F                      ;If mouse is shown,
                call    near ptr PutCursor      ;display cursor
@@:             popa                            ;Restore registers
                popf                            ;Restore flags
                retf                            ;Return

_MouseMove_     EndP

;**************************** PutCursor -- put cursor on screen, save screen

PutCursor       Proc    Near

                pusha
                push    ds

                push    0A000h                  ;DS = video memory
                pop     ds

                call    near ptr GetScrn        ;Save screen area

                mov     di,offset MouseCursor
                mov     bp,cs:[MouseX]
                mov     si,cs:[MouseY]
                mov     dx,13
pc_lnloop:      xor     bx,bx
                mov     cx,bp

;****************** This is the setpixel routine

pc_pxloop:      mov     al,cs:[di+bx]           ;Get cursor byte
                cmp     al,-1                   ;-1, skip this pixel
                je      sp_done

                cmp     cx,639                  ;X > 639?
                jg      pc_lnlb                 ;Next line
                cmp     si,399                  ;Y > 399?
                jg      pc_done

                pusha                           ;Save registers

                mov     bx,cx                   ;SI = y, BX = CX = x
                and     cl,3                    ;CL = bit plane
                mov     ch,al                   ;Save color in CH
                mov     ah,1                    ;AH = plane mask
                shl     ah,cl
                shr     bx,2                    ;BX = X offset

                mov     dx,03C4h                ;Sequencer port
                mov     al,2                    ;Set Map Mask
                out     dx,ax

                mov     ax,si                   ;Multiply Y by 160
                shl     si,2
                add     ax,si
                shl     ax,5
                add     bx,ax                   ;BX = offset
                mov     [bx],ch                 ;Put byte in memory

                popa                            ;Restore registers

sp_done:        inc     cx                      ;Next pixel
                inc     bx                      ;Inc counter
                cmp     bx,8
                jnz     pc_pxloop               ;Loop back

pc_lnlb:        inc     si                      ;Next line
                add     di,8
                dec     dx                      ;Dec counter
                jnz     pc_lnloop               ;Loop back

pc_done:        pop     ds
                popa
                ret

PutCursor       EndP

;**************************** GetScrn -- Save mouse cursor screen area

GetScrn         Proc    Near

                pusha                           ;Save registers

                mov     dx,03CEh                ;DX = Graphics port
                mov     cx,cs:[MouseX]          ;Get Mouse X position
                mov     di,cx                   ;also in DI
                and     cl,3                    ;AH = Plane value
                mov     ah,cl
                mov     al,4                    ;4 = Set Plane command

                shr     di,2                    ;DI = offset in row
                mov     bx,cs:[MouseY]          ;BX = Y position
                mov     si,bx                   ;SI = Y also
                shl     bx,2                    ;Multiply BX by 160
                add     bx,si
                shl     bx,5
                add     di,bx                   ;DI = offset

                mov     bp,4                    ;4 planes
                mov     bx,offset MouseBuffer   ;BX = buffer
gs_ploop:       out     dx,ax                   ;Send plane command

                push    ax                      ;Save AX, DX
                push    dx
                mov     dx,03DAh                ;Input Port 1

                mov     si,13                   ;13 rows
                push    di                      ;Save DI

gs_loop1:       in      al,dx                   ;Wait for HRT
                test    al,1
                jz      gs_loop1
gs_HRT1:        in      al,dx                   ;Wait for next HRT
                test    al,1
                jnz     gs_HRT1
                mov     cl,[di]                 ;Get first byte
gs_HRT2:        in      al,dx                   ;Wait for HRT
                test    al,1
                jz      gs_HRT2
gs_HRT3:        in      al,dx                   ;Wait for next HRT
                test    al,1
                jnz     gs_HRT3
                mov     ch,[di+1]               ;Get second byte
                mov     cs:[bx],cx              ;Store them
                add     di,160                  ;Next line
                add     bx,2
                dec     si                      ;Loop back
                jnz     gs_loop1
                pop     di                      ;Restore DI
                pop     dx                      ;Restore DX, AX
                pop     ax

                inc     ah
                cmp     ah,4
                jne     @F
                inc     di
                mov     ah,0

@@:             dec     bp
                jnz     gs_ploop

                popa
                ret

GetScrn         EndP

;**************************** PutScrn -- Restore cursor screen area

PutScrn         Proc    Near

                pusha                           ;Save registers
                push    ds
                push    0A000h                  ;DS = video memory
                pop     ds

                mov     dx,03C4h                ;DX = Sequencer port
                mov     cx,cs:[MouseX]          ;Get Mouse X position
                mov     di,cx                   ;also in DI
                and     cl,3                    ;CX = Plane number
                mov     ah,1                    ;AH = Plane value
                shl     ah,cl
                mov     al,2                    ;2 = Set Plane command

                shr     di,2                    ;DI = offset in row
                mov     bx,cs:[MouseY]          ;BX = Y position
                mov     si,bx                   ;SI = Y also
                shl     bx,2                    ;Multiply BX by 160
                add     bx,si
                shl     bx,5
                add     di,bx                   ;DI = offset

                mov     bp,4                    ;4 planes
                mov     bx,offset MouseBuffer   ;BX = buffer
ps_ploop:       out     dx,ax                   ;Send plane command

                mov     si,13                   ;13 rows
                push    di                      ;Save DI
ps_loop1:       mov     cx,cs:[bx]              ;Get 2 bytes
                mov     [di],cl                 ;Move first byte
                mov     [di+1],ch               ;Move second byte
                add     di,160                  ;Next line
                add     bx,2
                dec     si                      ;Loop back
                jnz     ps_loop1
                pop     di                      ;Restore DI

                add     ah,ah
                test    ah,10h
                jz      @F
                inc     di
                mov     ah,1

@@:             dec     bp
                jnz     ps_ploop

                pop     ds
                popa
                ret

PutScrn         EndP

MouseBuffer     db 104 dup(0)

MouseCursor     db -1, 0,-1,-1,-1,-1,-1,-1
                db  0,15, 0,-1,-1,-1,-1,-1
                db  0,15,15, 0,-1,-1,-1,-1
                db  0,15,15,15, 0,-1,-1,-1
                db  0,15,15,15,15, 0,-1,-1
                db  0,15,15,15,15,15, 0,-1
                db  0,15,15,15,15,15,15, 0
                db  0,15,15,15,15, 0, 0,-1
                db  0,15,15, 0,15,15, 0,-1
                db  0,15, 0, 0,15,15, 0,-1
                db -1, 0,-1,-1, 0,15,15, 0
                db -1,-1,-1,-1, 0,15,15, 0
                db -1,-1,-1,-1,-1, 0, 0,-1

End
